-- This file is part of KDevelop.
-- Copyright (c) 2005 Jakob Petsovits <jpetso@gmx.at>
--
-- This grammar is free software; you can redistribute it and/or
-- modify it under the terms of the GNU Lesser General Public
-- License as published by the Free Software Foundation; either
-- version 2.1 of the License, or (at your option) any later version.
--
-- This grammar is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
--
-- You should have received a copy of the GNU Lesser General Public License
-- along with this library; see the file COPYING.LIB.  If not, write to
-- the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-- Boston, MA 02110-1301, USA.


-- Grammar for Java 1.5
-- Modelled after the public domain ANTLR Java 1.5 grammar at
-- http://www.antlr.org/grammar/1090713067533/index.html
-- (Version 1.22.5 from January 03, 2005)
-- and the reference grammar of the Java language specification,
-- Third Edition, at http://java.sun.com/docs/books/jls/.



-- Known problematic conflicts (1 conflict):
--  - The first/follow IMPORT, STATIC, CLASS, INTERFACE etc. conflict
--    in compilation_unit (package_declaration vs. type_declaration).
--    This one makes either annotated packages or annotated classes impossible
--    when there is no import statement. I decided to have annotated classes.
--    (1 conflict)

-- Known harmless or resolved conflicts (17 conflicts):
--  - The first/follow AT and [IMPORT, AT] conflicts in compilation_unit
--    (manually resolved, 2 conflicts)
--  - The first/follow EXTENDS conflict in type_parameter,
--    the first/follow COMMA conflict in type_arguments,
--    the first/follow >, >>, >>> conflict in type_arguments,
--    the first/follow EXTENDS conflict in wildcard_type, and
--    the first/follow LBRACKET conflict in optional_declarator_brackets
--    (the latter one coming from class_type_declaration).
--    Seems to be the same reason, as there might be the same character in
--    some higher-level rule. I believe they are harmless, as the default
--    behaviour of being greedy should be the right thing here.
--    (5 conflicts)
--  - The first/first IDENTIFIER conflict in class_field
--    (manually resolved, 1 conflict)
--  - The first/first STATIC conflict in class_field
--    (manually resolved, 1 conflict)
--  - The first/first IDENTIFIER conflicts in *_field,
--    between method_name and variable_name (manually resolved, 4 conflicts)
--  - The first/first "0" conflicts in parameter_declaration_list
--    (done right by default, 2 conflicts)
--  - The first/first "0" conflicts in [non_wildcard_]type_arguments, which
--    is the same kind of conflicts like in parameter_declaration_list above
--    (from the main file, done right by default, 2 conflicts)

-- Conflicts of unknown consequence (6 conflicts):
--  - The first/first "0" conflicts in compilation_unit. Should be
--    harmless, but who knows? Roberto, is this an "empty-file case"
--    or also possible at other occasions? (2 conflicts)
--  - The first/first "0" conflict in enum_body, similar to the above
--    (1 conflict)
--  - The first/follow COMMA conflicts in enum_body
--    (2 conflicts)
--  - The first/follow IDENTIFIER conflict in parameter_declaration_list
--    (1 conflict)

-- Total amount of conflicts: 24



-- Global variables and initializations

-- Stores the amount of currently open type arguments rules,
-- all of which are beginning with a less than ("<") character.
-- This way, also RSIGNEDSHIFT (">>") and RUNSIGNEDSHIFT (">>>") can be used
-- to close type arguments rules, in addition to GREATERTHAN (">").
-- I'd rather have this as a member of the java class,
-- but didn't find out how to do this (if possible at all).
[: static int ltCounter = 0;  :]



-- The grammar starts with compilation_unit,
-- which is equivalent to a single source file.

   ( ?[: /* yytoken == Token_AT || */ yytoken == Token_PACKAGE :]
     -- type_declaration can also start with annotations. This condition is
     -- an incorrect solution to the conflict, the right one would only be
     -- solvable with LL(k) and look like
     -- ?[: LA( package_declaration_lookahead ) :] -- not supported by kdev-pg
      package_declaration=package_declaration
    | 0
   )
   ( ?[: yytoken == Token_IMPORT :] !(#import_declaration=import_declaration)
    | 0
   )
   !(#type_declaration=type_declaration)
-> compilation_unit ;;

--    !(#annotation=annotation) PACKAGE
-- -> package_declaration_lookahead ;;  -- only use for lookaheads!



-- A PACKAGE DECLARATION: optional annotations followed by "package",
--                        then the package identifier.
-- 1 first/first conflict, because annotation as well as "| 0"
-- inside of compilation_unit may both be 0. The ANTLR grammar
-- checks on ?[:annotations "package":] to do a package_declaration.

   !(#annotation=annotation) PACKAGE package_name=qualified_identifier SEMICOLON
-> package_declaration ;;


-- An IMPORT DECLARATION is "import" followed by a package or type (=class) name.

   IMPORT (token_static=STATIC | 0)
   identifier_name=qualified_identifier (DOT star=STAR | 0) SEMICOLON
-> import_declaration ;;


-- A TYPE DECLARATION is either a class, interface, enum or annotation.

   !(#modifier=modifier)
   (  class_declaration=class_declaration
    | enum_declaration=enum_declaration
    | interface_declaration=interface_declaration
    | annotation_type_declaration=annotation_type_declaration
   )
   SEMICOLON   -- the ANTLR grammar makes multiple semicolons possible
               -- although it's not in the Java specification
-> type_declaration ;;



-- Definition of a Java CLASS

   CLASS class_name=IDENTIFIER
   (type_parameters=type_parameters  | 0)  -- it might have type parameters
   (extends=class_extends_clause     | 0)  -- it might have a super class
   (implements=implements_clause     | 0)  -- it might implement some interfaces
   body=class_body
-> class_declaration ;;


-- Definition of a Java INTERFACE

   INTERFACE interface_name=IDENTIFIER
   (type_parameters=type_parameters  | 0)  -- it might have type parameters
   (extends=interface_extends_clause | 0)  -- it might extend other interfaces
   body=interface_body
-> interface_declaration ;;


-- Definition of ENUMERATIONs and ANNOTATION TYPEs

   ENUM enum_name=IDENTIFIER
   (implements=implements_clause     | 0)  -- it might implement some interfaces
   body=enum_body
-> enum_declaration ;;

   AT INTERFACE annotation_type_name=IDENTIFIER
   body=annotation_type_body
-> annotation_type_declaration ;;



-- BODIES of classes, interfaces, annotation types and enums.

   LBRACE !(#declaration=class_field) RBRACE
-> class_body ;;

   LBRACE !(#declaration=interface_field) RBRACE
-> interface_body ;;

   LBRACE !(#annotation_type_field=annotation_type_field) RBRACE
-> annotation_type_body ;;

-- In an enum body, you can have zero or more enum constants
-- followed by any number of fields like a regular class.

   LBRACE
   ( #enum_constant=enum_constant
     !(  ?[: (LA(2).kind != Token_SEMICOLON && LA(2).kind != Token_RBRACE) :]
         ( COMMA #enum_constant=enum_constant )
       | 0 [: break; /* exit the while loop and go ahead */ :]
      )
   | 0 )
   (COMMA | 0)
   ( SEMICOLON !(#class_field=class_field) | 0 )
   RBRACE
-> enum_body ;;

-- An enum constant may have optional parameters and may have a class body

   !( #annotation=annotation ) identifier=IDENTIFIER
   (LPAREN arguments=argument_list RPAREN | 0)
   (body=enum_constant_body | 0)
-> enum_constant ;;

   LBRACE !(#declaration=enum_constant_field) RBRACE
-> enum_constant_body ;;



-- Short CLAUSES used in various declarations

   EXTENDS type=class_or_interface_type
-> class_extends_clause ;;

   EXTENDS #type=class_or_interface_type @ COMMA
-> interface_extends_clause ;;

   IMPLEMENTS #type=class_or_interface_type @ COMMA
-> implements_clause ;;

   THROWS #identifier=qualified_identifier @ COMMA
-> throws_clause ;;




-- Now on to what happens inside the class, interface, etc. bodies:


-- An ANNOTATION TYPE FIELD. The Java specification has a bug here,
-- two different AnnotationTypeElementDeclarations in the book's body
-- and the grammar appendix. I chose the one from the body,
-- which is the same that the ANTLR grammar also uses.

 ( !(#modifier=modifier)
   (  class_declaration=class_declaration
    | enum_declaration=enum_declaration
    | interface_declaration=interface_declaration
    | annotation_type_declaration=annotation_type_declaration
    |
      type=type_specification
      (                      -- annotation method without arguments:
         ?[: LA(2).kind == Token_LPAREN :] -- resolves the IDENTIFIER conflict
                                      -- between method_name and variable_name
         identifier=IDENTIFIER
         LPAREN RPAREN
         -- declarator_brackets=optional_declarator_brackets -- ANTLR grammar's bug:
         -- It's not in the Java Spec, and obviously has been copied
         -- from classField even if it doesn't belong here.
         (DEFAULT annotation_element_value=annotation_element_value | 0)
         SEMICOLON
       |                     -- or a ConstantDeclaration:
         #variable_declarator=variable_declarator @ COMMA
         SEMICOLON
      )
   )
 | SEMICOLON
 )
-> annotation_type_field ;;


-- A CLASS FIELD, representing the various things
-- that can be defined inside a class.

 ( ?[: LA(2).kind != Token_LBRACE :] -- resolves the conflict, see further down
   !(#modifier=modifier)
   (  class_declaration=class_declaration
    | enum_declaration=enum_declaration
    | interface_declaration=interface_declaration
    | annotation_type_declaration=annotation_type_declaration
    |
      -- A generic method/ctor has the type_parameters before the return type.
      -- This is not allowed for variable definitions, but this production
      -- allows it, a semantic check could be used if you wanted.
      -- (Note: kdev-pg does not yet support semantic checks.)
      (type_params=type_parameters | 0)
      (
         -- constructor declaration (without prepended type_specification)
         ?[: LA(2).kind == Token_LPAREN :]
         -- resolves the IDENTIFIER conflict with type_specification
         constructor_name=IDENTIFIER
         constructor_parameters=parameter_declaration_list
         (constructor_throws_clause=throws_clause | 0)
         constructor_body=block
         -- leaving out explicit this(...) and super(...) invocations,
         -- these are just normal statements for the grammar
       |
         -- method or variable declaration
         type=type_specification
         (
            ?[: LA(2).kind == Token_LPAREN :] -- resolves the IDENTIFIER
                          -- conflict between method_name and variable_name
            method_name=IDENTIFIER
            method_parameters=parameter_declaration_list
            method_declarator_brackets=optional_declarator_brackets
            (method_throws_clause=throws_clause | 0)
            (method_body=block | SEMICOLON)
          |
            #variable_declarator=variable_declarator @ COMMA
            SEMICOLON
         )
      )
   )
 | -- The static class initializer block ("static {...}") has a conflict with
   -- the modifiers from above, which can also be static. A block must always
   -- start with "{", so when encountering static, this can be used to resolve
   -- this conflict.
   ?[: LA(2).kind == Token_LBRACE :] STATIC static_initializer_block=block
 | instance_initializer_block=block  -- "{...}" instance initializer
 | SEMICOLON
 )
-> class_field ;;


-- An ENUM CONSTANT FIELD is just like a class field but without
-- the possibility of a constructor definition or a static initializer.

 ( !(#modifier=modifier)
   (  class_declaration=class_declaration
    | enum_declaration=enum_declaration
    | interface_declaration=interface_declaration
    | annotation_type_declaration=annotation_type_declaration
    |
      -- A generic method/ctor has the type_parameters before the return type.
      -- This is not allowed for variable definitions, but this production
      -- allows it, a semantic check could be used if you wanted.
      -- (Note: kdev-pg does not yet support semantic checks.)
      (type_params=type_parameters | 0) type=type_specification
      (
         ?[: LA(2).kind == Token_LPAREN :] -- resolves the IDENTIFIER conflict
                                      -- between method_name and variable_name
         method_name=IDENTIFIER
         method_parameters=parameter_declaration_list
         method_declarator_brackets=optional_declarator_brackets
         (method_throws_clause=throws_clause | 0)
         (method_body=block | SEMICOLON)
       |
         #variable_declarator=variable_declarator @ COMMA
         SEMICOLON
      )
   )
 | instance_initializer_block=block  -- "{...}" instance initializer
 | SEMICOLON
 )
-> enum_constant_field ;;


-- An INTERFACE FIELD is the same as an enum constant field but without
-- the possibility of any initializers or method blocks.

 ( !(#modifier=modifier)
   (  class_declaration=class_declaration
    | enum_declaration=enum_declaration
    | interface_declaration=interface_declaration
    | annotation_type_declaration=annotation_type_declaration
    |
      -- A generic method/ctor has the type_parameters before the return type.
      -- This is not allowed for variable definitions, but this production
      -- allows it, a semantic check could be used if you wanted.
      -- (Note: kdev-pg does not yet support semantic checks.)
      (type_params=type_parameters | 0) type=type_specification
      (
         ?[: LA(2).kind == Token_LPAREN :] -- resolves the IDENTIFIER conflict
                                      -- between method_name and variable_name
         method_name=IDENTIFIER
         method_parameters=parameter_declaration_list
         method_declarator_brackets=optional_declarator_brackets
         (method_throws_clause=throws_clause | 0)
         SEMICOLON
       |
         #variable_declarator=variable_declarator @ COMMA
         SEMICOLON
      )
   )
 | SEMICOLON
 )
-> interface_field ;;




-- So much for the rough structure, now we get into the details


-- A PARAMETER DECLARATION LIST is part of a method header and can contain
-- zero or more parameters, optionally ending with a variable-length parameter.
-- Oh yeah, and you'll notice that this is a big hack, which is only there
-- because kdev-pg doesn't offer exit conditions from a !(...) or "@" operator.

   LPAREN [: bool may_proceed = true; :]
   (
      #parameter_modifiers=parameter_modifiers
      #type_specification=type_specification
      ( triple_dot=TRIPLEDOT [: may_proceed = false; :] | 0 )
      #variable_identifier=IDENTIFIER
      #declarator_brackets=optional_declarator_brackets
      (
         ?[: may_proceed == true :]
         !(
            COMMA
            #parameter_modifiers=parameter_modifiers
            #type_specification=type_specification
            ( triple_dot=TRIPLEDOT [: may_proceed = false; :] | 0 )
            #variable_identifier=IDENTIFIER
            #declarator_brackets=optional_declarator_brackets
            [: if( may_proceed == false ) { break; } :]
            -- Don't proceed after the triple dot. If there's a sane way
            -- to exit the loop when may_proceed==false,
            -- please use that instead of this whole nasty construct.
         )
       | 0  -- when we get here: one argument
      )
    | 0  -- when we get here: no arguments
   )
   RPAREN
-> parameter_declaration_list ;;

-- How it _should_ look:
--
--    LPAREN [: bool may_proceed = true; :]
--    (
--       (
--          #parameter_modifiers=parameter_modifiers
--          #type_specification=type_specification
--          ( triple_dot=TRIPLEDOT [: may_proceed = false; :] | 0 )
--          #variable_identifier=IDENTIFIER
--          #declarator_brackets=optional_declarator_brackets
--       )
--       @ ( ?[: may_proceed == true :] COMMA )
--           -- kdev-pg dismisses this condition!
--     |
--       0
--    )
--    RPAREN
-- -> parameter_declaration_list ;;

   !( token_final=FINAL | #annotation=annotation )
-> parameter_modifiers ;;

-- The PARAMETER DECLARATION rule is not used in parameter_declaration_list,
-- because otherwise the may_proceed approach couldn't be used there.
-- It is used in the try_handler.

   #parameter_modifiers=parameter_modifiers
   #type_specification=type_specification
   #variable_identifier=IDENTIFIER
   #declarator_brackets=optional_declarator_brackets
-> parameter_declaration ;;



-- The body of a CONSTRUCTOR METHOD is the same as a normal block, except that
-- it's possible to have a constructor call like this(...) or super(...)
-- at the beginning of the block. This causes a conflict which is difficult
-- to resolve, so class_field uses block instead of constructor_body.
-- TODO: make sure that the THIS and SUPER stuff is in statement too,
--       as this commented stuff can't provide it.
--
--    LBRACE
--    (explicit_constructor_invocation=explicit_constructor_invocation | 0)
--    !(#statement=statement)
--    RBRACE
-- -> constructor_body ;;
--
-- -- Catches obvious constructor calls, but not the expr.super(...) calls:
--
--    (type_arguments=type_arguments | 0)
--    (invocated_constructor=THIS | invocated_constructor=SUPER)
--    LPAREN arguments=argument_list RPAREN SEMICOLON
-- -> explicit_constructor_invocation ;;



-- VARIABLE DECLARATIONS, initializers, etc.

-- A VARIABLE DECLARATOR, as used in a class_field, enum_field, etc.

   variable_name=IDENTIFIER
   declarator_brackets=optional_declarator_brackets
   ( ASSIGN initializer=variable_initializer | 0 )
-> variable_declarator ;;

   ( expression=expression | array_initializer=variable_array_initializer )
-> variable_initializer ;;

   LBRACE
   (  (#variable_initializer=variable_initializer @ COMMA (COMMA | 0))
    | 0
   )
   RBRACE
-> variable_array_initializer ;;




-- Type parameters and type arguments, the two rules responsible for the
-- greater-than special casing. (This is the generic aspect in Java >= 1.5.)

-- TYPE PARAMETERS are used in class, interface etc. declarations to
-- determine the generic types allowed as type argument.

   LT [: int currentLtLevel = ltCounter; ltCounter++; :]
   #type_parameter=type_parameter @ COMMA
   (
      type_arguments_or_parameters_end
      -- make sure we have gobbled up enough '>' characters
      -- if we are at the "top level" of nested type_parameters productions
      [: if( currentLtLevel == 0 && ltCounter != currentLtLevel )
          { return false; }
       :]
      -- TODO: error message, saying that the amount of ">" chars is incorrect
    |
      0  -- they can also be changed by type_parameter or type_argument
   )
-> type_parameters ;;

   identifier=IDENTIFIER
   (EXTENDS (#type=class_or_interface_type @ BIT_AND) | 0)
-> type_parameter ;;


-- TYPE ARGUMENTS are used in initializers, invocations, etc. to
-- specify the exact types for this generic class/method instance.

   LT [: int currentLtLevel = ltCounter; ltCounter++; :]
   #type_argument=type_argument
   (
      ?[: ltCounter == currentLtLevel + 1 :]
      !( COMMA #type_argument=type_argument
         [: if( ltCounter != currentLtLevel + 1 ) break; :]
         -- hack, see parameter_declaration_list
      )
    |
      0
   )
   (
      type_arguments_or_parameters_end
      -- make sure we have gobbled up enough '>' characters
      -- if we are at the "top level" of nested type_argument productions
      [: if( currentLtLevel == 0 && ltCounter != currentLtLevel )
          { return false; }
       :]
      -- TODO: error message, saying that the amount of ">" chars is incorrect
    |
      0  -- they can also be changed by type_parameter or type_argument
   )
-> type_arguments ;;

   LT [: int currentLtLevel = ltCounter; ltCounter++; :]
   #type_argument_specification=type_argument_specification
   (
      ?[: ltCounter == currentLtLevel + 1 :]
      !( COMMA #type_argument=type_argument
         [: if( ltCounter != currentLtLevel + 1 ) break; :]
         -- hack, see parameter_declaration_list
      )
    |
      0
   )
   (
      type_arguments_or_parameters_end
      -- make sure we have gobbled up enough '>' characters
      -- if we are at the "top level" of nested type_argument productions
      [: if( currentLtLevel == 0 && ltCounter != currentLtLevel )
          { return false; }
       :]
      -- TODO: error message, saying that the amount of ">" chars is incorrect
    |
      0  -- they can also be changed by type_parameter or type_argument
   )
-> non_wildcard_type_arguments ;;

   type_argument_specification=type_argument_specification
 | wildcard_type=wildcard_type
-> type_argument ;;

-- Any type specification except primitive types is allowed as type argument.
-- Arrays of primitive types are allowed nonetheless.

   class_type=class_type_specification
 | builtin_type_array=builtin_type_array_specification
-> type_argument_specification ;;

   QUESTION (bounds=wildcard_type_bounds | 0)
-> wildcard_type ;;

   (extends_or_super=EXTENDS | extends_or_super=SUPER)
   type=class_or_interface_type
-> wildcard_type_bounds ;;


   GREATERTHAN    [: ltCounter -= 1; :]  -- ">"
 | RSIGNEDSHIFT   [: ltCounter -= 2; :]  -- ">>"
 | RUNSIGNEDSHIFT [: ltCounter -= 3; :]  -- ">>>"
-> type_arguments_or_parameters_end ;;




-- All kinds of rules for types here.

-- A TYPE SPECIFICATION is a type name with possible brackets afterwards
-- (which would make it an array type). Called "Type" in the Java spec.

   class_type_spec=class_type_specification
 | builtin_type_spec=builtin_type_specification
-> type_specification ;;

   type=class_or_interface_type
   declarator_brackets=optional_declarator_brackets
-> class_type_specification ;;

   type=builtin_type declarator_brackets=optional_declarator_brackets
-> builtin_type_specification ;;

   type=builtin_type declarator_brackets=mandatory_declarator_brackets
-> builtin_type_array_specification ;;


-- A SIMPLE TYPE SPECIFICATION is just a type name,
-- with no possibility of brackets afterwards.

   class_or_interface_type=class_or_interface_type
 | builtin_type=builtin_type
-> type_specification_noarray ;;

-- The primitive types. The Java specification doesn't include void here,
-- but the ANTLR grammar works that way.

   type=VOID
 | type=BOOLEAN
 | type=BYTE
 | type=CHAR
 | type=SHORT
 | type=INT
 | type=FLOAT
 | type=LONG
 | type=DOUBLE
-> builtin_type ;;

   #part=class_or_interface_type_part @ DOT
-> class_or_interface_type ;;

   identifier=IDENTIFIER (type_arguments=type_arguments | 0)
-> class_or_interface_type_part ;;



-- QUALIFIED IDENTIFIERs are either qualified ones or raw identifiers.

   #name=IDENTIFIER @ DOT
-> qualified_identifier ;;

-- Declarator brackets are part of a type specification, like String[][]
-- They are always empty, only have to be counted.

   !( #lbracket=LBRACKET RBRACKET )
-> optional_declarator_brackets ;;
-- TODO: make a counter instead of filling the sequence array with lbrackets

   #lbracket=LBRACKET RBRACKET
   !( #lbracket=LBRACKET RBRACKET )
-> mandatory_declarator_brackets ;;
-- TODO: make a counter instead of filling the sequence array with lbrackets




-- ANNOTATIONS look for example like @Info( name="Jakob", born=1983 ),
-- or @Info("Jakob"), or just @Info, and are attached to a method, class or package.
-- @Info is equivalent to @Info().

   AT type_name=qualified_identifier
   ( LPAREN  (args=annotation_arguments | 0) RPAREN  | 0 )
-> annotation ;;

 ( ( ?[: LA(2).kind == Token_ASSIGN :]
     #value_pair=annotation_element_value_pair @ COMMA
   )
 | element_value=annotation_element_value  -- element_name is "value" here
 )
-> annotation_arguments ;;

   element_name=IDENTIFIER ASSIGN element_value=annotation_element_value
-> annotation_element_value_pair ;;

   cond_expression=conditional_expression
 | annotation=annotation
 | element_array_initializer=annotation_element_array_initializer
-> annotation_element_value ;;

-- Same as annotation_element_value, but array_initializer is excluded.
-- That's because nested annotation array initialisers are not valid.
-- (The Java specification hides that in a short "discussion" area.)
   cond_expression=conditional_expression
 | annotation=annotation
-> annotation_element_array_value ;;

   LBRACE
   !(#element_value=annotation_element_array_value @ COMMA) (COMMA | 0)
   RBRACE
-> annotation_element_array_initializer ;;




-- And now for the good stuff: statements, expressions and the likes. Yay!

-- This is a BLOCK, a list of statements. It is used in many contexts:
-- Inside a class definition prefixed with "static" as class initializer
-- Inside a class definition without "static" as instance initializer
-- As the body of a method
-- As a completely independent braced block of code inside a method,
--  starting a new scope for variable definitions

   LBRACE !(#statement=statement) RBRACE
-> block ;;

-- -> statement ;;



-- An ARGUMENT LIST mostly occurs in method calls, always between an LPAREN
-- and an RPAREN, to specify the values for the called method/initializer/etc.

   #expression=expression @ COMMA
 | 0
-> argument_list ;;




-- MODIFIERs for Java classes, interfaces, class/instance vars and methods.
-- Sometimes not all of them are valid, but that has to be checked manually
-- after running the parser. The ANTLR grammar also does it this way.

   token=PRIVATE
 | token=PUBLIC
 | token=PROTECTED
 | token=STATIC
 | token=TRANSIENT
 | token=FINAL
 | token=ABSTRACT
 | token=NATIVE
 -- Neither in the Java spec nor in the JavaCC grammar, just in the ANTLR one:
 -- | token=THREADSAFE
 | token=SYNCHRONIZED
 | token=VOLATILE
 | token=STRICTFP
 -- A modifier may be any annotation (e.g. @bla), but not @interface.
 -- This condition resolves the conflict between modifiers
 -- and annotation type declarations:
 | ?[: yytoken == Token_AT && LA(2).kind != Token_INTERFACE :]
    annotation=annotation
-> modifier ;;





--
-- Appendix 1: stubs, so that the (incomplete) parser can be created
--

   STUB_A -> conditional_expression ;;
   STUB_B -> statement ;;
   STUB_C -> expression ;;


--
-- Appendix 2: token list, so they will be defined in java.h
--

   ( keyword | seperator | operator | literal | IDENTIFIER )
-> token ;;

   (  ABSTRACT | ASSERT | BOOLEAN | BREAK | BYTE | CASE | CATCH | CHAR | CLASS | CONST | CONTINUE
    | DEFAULT | DO | DOUBLE | ELSE | ENUM | EXTENDS | FINAL | FINALLY | FLOAT | FOR | GOTO
    | IF | IMPLEMENTS | IMPORT | INSTANCEOF | INT | INTERFACE | LONG | NATIVE | NEW | PACKAGE
    | PRIVATE | PROTECTED | PUBLIC | RETURN | SHORT | STATIC | STRICTFP | SUPER | SWITCH
    | SYNCHRONIZED | THIS | THROW | THROWS | TRANSIENT | TRY | VOID | VOLATILE | WHILE )
-> keyword ;;

   ( LPAREN | RPAREN | LBRACE | RBRACE | LBRACKET | RBRACKET | SEMICOLON | COMMA | DOT | AT )
-> seperator ;;

   (  ASSIGN | LESSTHAN | BANG | TILDE | QUESTION | COLON | EQUAL | LESSEQUAL | GREATEREQUAL
    | NOTEQUAL | LOG_AND | LOG_OR | INCREMENT | DECREMENT | PLUS | MINUS | STAR | SLASH | BIT_AND
    | BIT_OR | XOR | REMAINDER | LSHIFT | PLUSASSIGN | MINUSASSIGN | STARASSIGN | SLASHASSIGN
    | ANDASSIGN | ORASSIGN | XORASSIGN | REMAINDERASSIGN | LSHIFTASSIGN | RSIGNEDSHIFTASSIGN
    | RUNSIGNEDSHIFTASSIGN | GREATERTHAN | RSIGNEDSHIFT | RUNSIGNEDSHIFT | TRIPLEDOT )
-> operator ;;

   ( boolean_literal | NULL | INTEGER_LITERAL | FLOATING_POINT_LITERAL | CHARACTER_LITERAL | STRING_LITERAL )
-> literal ;;

   ( value=FALSE | value=TRUE )
-> boolean_literal ;;
