-- This file is part of KDevelop.
-- Copyright (c) 2005 Jakob Petsovits <jpetso@gmx.at>
--
-- This grammar is free software; you can redistribute it and/or
-- modify it under the terms of the GNU Lesser General Public
-- License as published by the Free Software Foundation; either
-- version 2.1 of the License, or (at your option) any later version.
--
-- This grammar is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
--
-- You should have received a copy of the GNU Lesser General Public License
-- along with this library; see the file COPYING.LIB.  If not, write to
-- the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-- Boston, MA 02110-1301, USA.


-- Grammar for Java 1.5
-- Modelled after the public domain ANTLR Java 1.5 grammar at
-- http://www.antlr.org/grammar/1090713067533/index.html
-- (Version 1.22.5 from January 03, 2005)
-- and the reference grammar of the Java language specification,
-- Third Edition, at http://java.sun.com/docs/books/jls/.



-- Known problematic conflicts (4 conflicts), requiring automatic LL(k):
--  - The first/first IDENTIFIER, VOID, BOOLEAN etc. conflict in block_statement
--    which is expression statements vs. variable declarations.
--    This one makes variable declarations impossible.
--    (You know, variable declarations can also start with IDENTIFIER @ DOT.)
--    The IDENTIFIER conflict in the same error message caused by
--    labeled statements is independent though, this one has been resolved.
--    The SYNCHRONIZED conflict in this error message is also
--    independent and has been resolved too.
--    (1 conflict)
--  - The first/first IDENTIFIER, FINAL, 0, VOID etc. conflict in for_control.
--    This one makes for-each clauses impossible, in favour of traditional
--    for-clauses. The "0" should be harmless though.
--    (1 conflict)
--  - The first/first IDENTIFIER, VOID, BOOLEAN etc. conflict in for_init.
--    This one makes expression lists in the for initializer impossible,
--    in favour of variable declarations.
--    (1 conflict)
--  - The first/first LPAREN conflict in unary_expression_not_plusminus
--    which is cast_expression vs. primary_expression.
--    This one makes casting impossible. Casting to builtin types would be
--    possible without LL(k), but I rather stick to a readable grammar.
--    (1 conflict)

-- Known harmless or resolved conflicts (16 conflicts):
--  - The first/follow COMMA conflict in type_arguments,
--    the first/follow >, >>, >>> conflict in type_arguments, and
--    the first/follow LBRACKET conflict in optional_declarator_brackets
--    (the latter one coming from class_type_declaration).
--    Seems to be the same reason, as there might be the same character in
--    some higher-level rule. I believe they are harmless, as the default
--    behaviour of being greedy should be the right thing here.
--    (from the main file, 3 conflicts)
--  - The first/follow ELSE conflict in expression, better known as
--    "dangling else" problem (done right by default, 1 conflict)
--  - The first/follow DOT, LBRACKET (+CLASS) conflict in primary_atom
--    (done right by default, 1 conflict)
--  - The first/follow LBRACE (class_body) conflict in new_expression
--    (done right by default, 1 conflict)
--  - The first/follow LBRACKET conflict in array_creator_rest
--    (done right by default, 1 conflict)
--  - The first/first "0" conflicts within [non_wildcard_]type_arguments
--    (from the main file, done right by default, 2 conflicts)
--  - The first/first IDENTIFIER conflict in annotation_arguments
--    (manually resolved, 1 conflict)
--  - The first/first IDENTIFIER conflict in statement
--    (manually resolved, 1 conflict)
--  - The first/first "0" conflict in for_clause
--    (done right by default, 1 conflict)
--  - The first/first IDENTIFIER conflicts in primary_selector and
--    super_suffix. Could be written without conflict, but done on purpose
--    to tell methods (with possible type arguments) and variables
--    (without these) apart. (2 identical conflicts)
--  - The first/follow LBRACKET conflicts in array_creator_rest.
--    This is by design and works as expected.
--    (manually resolved, 2 conflicts)

-- Conflicts of unknown consequence (11 conflicts):
--  - The first/follow SUPER conflict in wildcard_type (coming from
--    class_or_interface_type_part). There must be some possibility to get
--    a conflict between wildcard_type_bounds and primary_expression here.
--    Being greedy should do well though, I think. (1 conflict)
--  - The first/first "0" conflict in primary_atom. Don't ask me what it
--    means and why it's present. (1 conflict)
--  - First/follow conflicts in expression. The expressions are greedy,
--    so the correct code should be produced, BUT if you look for example
--    at parse_argument_list(), you'll notice that too much is matched.
--    I would bet on the bang bug. The conflicts are:
--      PLUS, MINUS, INCREMENT, DECREMENT (coming from statement)
--      PLUS, MINUS, INCREMENT, DECREMENT (x2, coming from for_clause)
--      PLUS, MINUS, INCREMENT, DECREMENT (coming from for_init)
--      ASSIGN, PLUSASSIGN, MINUSASSIGN, etc. (coming from expression)
--      QUESTION (coming from conditional_expression)
--      GREATERTHAN, INSTANCEOF, LESSTHAN, etc. (...relational_expression)
--      PLUS, MINUS, INCREMENT, DECREMENT (x2, coming from argument_list)
--    (9 conflicts)

-- Total amount of conflicts: 31, from which 5 are coming from the main file



-- A pseudo start rule.

   !(statement)
-> compilation_unit ;;

-- A pseudo class body.

   LBRACE !(#declaration=STUB_X) RBRACE
-> class_body ;;



-- The PARAMETER DECLARATION rule is not used in parameter_declaration_list,
-- because otherwise the may_proceed approach couldn't be used there.
-- It is used in the try_handler.

   #parameter_modifiers=parameter_modifiers
   #type_specification=type_specification
   #variable_identifier=IDENTIFIER
   #declarator_brackets=optional_declarator_brackets
-> parameter_declaration ;;

   !( token_final=FINAL | #annotation=annotation )
-> parameter_modifiers ;;



-- VARIABLE DECLARATIONS, initializers, etc.

-- A VARIABLE DECLARATOR, as used within a class_field, enum_field, etc.

   variable_name=IDENTIFIER
   declarator_brackets=optional_declarator_brackets
   ( ASSIGN initializer=variable_initializer | 0 )
-> variable_declarator ;;

   ( expression=expression | array_initializer=variable_array_initializer )
-> variable_initializer ;;

   LBRACE
   (  (#variable_initializer=variable_initializer @ COMMA (COMMA | 0))
    | 0
   )
   RBRACE
-> variable_array_initializer ;;



-- TYPE ARGUMENTS are used in initializers, invocations, etc. to
-- specify the exact types for this generic class/method instance.

   LT [: int currentLtLevel = ltCounter; ltCounter++; :]
   #type_argument=type_argument
   (
      ?[: ltCounter == currentLtLevel + 1 :]
      !( COMMA #type_argument=type_argument
         [: if( ltCounter != currentLtLevel + 1 ) break; :]
         -- hack, see parameter_declaration_list
      )
    |
      0
   )
   (
      type_arguments_or_parameters_end
      -- make sure we have gobbled up enough '>' characters
      -- if we are at the "top level" of nested type_argument productions
      [: if( currentLtLevel == 0 && lessThanCounter != currentLtLevel )
          { return false; }
       :]
      -- TODO: error message, saying that the amount of ">" chars is incorrect
    |
      0  -- they can also be changed by type_parameter or type_argument
   )
-> type_arguments ;;

   LT [: int currentLtLevel = lessThanCounter; lessThanCounter++; :]
   #type_argument_specification=type_argument_specification
   (
      ?[: lessThanCounter == currentLtLevel + 1 :]
      !( COMMA #type_argument=type_argument
         [: if( lessThanCounter != currentLtLevel + 1 ) break; :]
         -- hack, see parameter_declaration_list
      )
    |
      0
   )
   (
      type_arguments_or_parameters_end
      -- make sure we have gobbled up enough '>' characters
      -- if we are at the "top level" of nested type_argument productions
      [: if( currentLtLevel == 0 && lessThanCounter != currentLtLevel )
          { return false; }
       :]
      -- TODO: error message, saying that the amount of ">" chars is incorrect
    |
      0  -- they can also be changed by type_parameter or type_argument
   )
-> non_wildcard_type_arguments ;;

   type_argument_specification=type_argument_specification
 | wildcard_type=wildcard_type
-> type_argument ;;

-- Any type specification except primitive types is allowed as type argument.
-- Arrays of primitive types are allowed nonetheless.

   class_type=class_type_specification
 | builtin_type_array=builtin_type_array_specification
-> type_argument_specification ;;

   QUESTION (bounds=wildcard_type_bounds | 0)
-> wildcard_type ;;

   (extends_or_super=EXTENDS | extends_or_super=SUPER)
   type=class_or_interface_type
-> wildcard_type_bounds ;;


   GREATERTHAN    [: lessThanCounter -= 1; :]  -- ">"
 | RSIGNEDSHIFT   [: lessThanCounter -= 2; :]  -- ">>"
 | RUNSIGNEDSHIFT [: lessThanCounter -= 3; :]  -- ">>>"
-> type_arguments_or_parameters_end ;;




-- All kinds of rules for types here.

-- A TYPE SPECIFICATION is a type name with possible brackets afterwards
-- (which would make it an array type). Called "Type" in the Java spec.

   class_type_spec=class_type_specification
 | builtin_type_spec=builtin_type_specification
-> type_specification ;;

   type=class_or_interface_type
   declarator_brackets=optional_declarator_brackets
-> class_type_specification ;;

   type=builtin_type declarator_brackets=optional_declarator_brackets
-> builtin_type_specification ;;

   type=builtin_type declarator_brackets=mandatory_declarator_brackets
-> builtin_type_array_specification ;;


-- A SIMPLE TYPE SPECIFICATION is just a type name,
-- with no possibility of brackets afterwards.

   class_or_interface_type=class_or_interface_type
 | builtin_type=builtin_type
-> type_specification_noarray ;;

-- The primitive types. The Java specification doesn't include void here,
-- but the ANTLR grammar works that way.

   type=VOID
 | type=BOOLEAN
 | type=BYTE
 | type=CHAR
 | type=SHORT
 | type=INT
 | type=FLOAT
 | type=LONG
 | type=DOUBLE
-> builtin_type ;;

   #part=class_or_interface_type_part @ DOT
-> class_or_interface_type ;;

   identifier=IDENTIFIER (type_arguments=type_arguments | 0)
-> class_or_interface_type_part ;;



-- QUALIFIED IDENTIFIERs are either qualified ones or raw identifiers.

   #name=IDENTIFIER @ DOT
-> qualified_identifier ;;

-- Declarator brackets are part of a type specification, like String[][]
-- They are always empty, only have to be counted.

   !( #lbracket=LBRACKET RBRACKET )
-> optional_declarator_brackets ;;
-- TODO: make a counter instead of filling the sequence array with lbrackets

   #lbracket=LBRACKET RBRACKET
   !( #lbracket=LBRACKET RBRACKET )
-> mandatory_declarator_brackets ;;
-- TODO: make a counter instead of filling the sequence array with lbrackets




-- ANNOTATIONS look for example like @Info( name="Jakob", born=1983 ),
-- or @Info("Jakob"), or just @Info, and are attached to a method, class or package.
-- @Info is equivalent to @Info().

   AT type_name=qualified_identifier
   ( LPAREN  (args=annotation_arguments | 0) RPAREN  | 0 )
-> annotation ;;

 ( ( ?[: LA(2).kind == Token_ASSIGN :]
     #value_pair=annotation_element_value_pair @ COMMA
   )
 | element_value=annotation_element_value  -- element_name is "value" here
 )
-> annotation_arguments ;;

   element_name=IDENTIFIER ASSIGN element_value=annotation_element_value
-> annotation_element_value_pair ;;

   cond_expression=conditional_expression
 | annotation=annotation
 | element_array_initializer=annotation_element_array_initializer
-> annotation_element_value ;;

-- Same as annotation_element_value, but array_initializer is excluded.
-- That's because nested annotation array initialisers are not valid.
-- (The Java specification hides that in a short "discussion" area.)
   cond_expression=conditional_expression
 | annotation=annotation
-> annotation_element_array_value ;;

   LBRACE
   !(#element_value=annotation_element_array_value @ COMMA) (COMMA | 0)
   RBRACE
-> annotation_element_array_initializer ;;




-- START COPYING TO THE ORIGINAL FILE FROM HERE ON




-- And now for the good stuff: statements, expressions and the likes. Yay!

-- This is a BLOCK, a list of statements. It is used in many contexts:
-- Inside a class definition prefixed with "static" as class initializer
-- Inside a class definition without "static" as instance initializer
-- As the body of a method
-- As a completely independent braced block of code inside a method,
--  starting a new scope for variable definitions

   LBRACE !(#statement=block_statement) RBRACE
-> block ;;

-- A BLOCK STATEMENT is either a normal statement, a variable declaration
-- or a type declaration (you know, nested classes and the likes...).
-- To avoid ambiguities, the variable declarations and type declarations
-- are handled together. But that doesn't resolve the other conflict.

 (
   ?[: (yytoken != Token_SYNCHRONIZED) ||
       (yytoken == Token_SYNCHRONIZED && LA(2).kind == Token_LPAREN)
     :]  -- resolves the SYNCHRONIZED conflict between statement and modifier.
   statement
 |
   -- This sub-rule unites two of the original rules in the Java spec's
   -- BlockStatement: The local variable declarations and type declarations.
   !(#modifier=modifier)
   (
      -- Variable declarations are ambiguous with "ID DOT" relative to
      -- expression statements. The ANTLR grammar uses a semantic check.
      -- TODO: this is gonna be a serious problem!
      -- TODO: when this can be LA()'d, move it in front of 'statement'
      variable_declaration_rest=variable_declaration_rest SEMICOLON
    |
      -- Inside a block, our four "complex types" can be declared
      -- (nested classes and the likes...):
      class_declaration=class_declaration
    | enum_declaration=enum_declaration
    | interface_declaration=interface_declaration
    | annotation_type_declaration=annotation_type_declaration
   )
 )
-> block_statement ;;

   !(#modifier=modifier) rest=variable_declaration_rest
-> variable_declaration ;;

   type_specification=type_specification
   #variable_declarator=variable_declarator @ COMMA
-> variable_declaration_rest ;;



-- The STATEMENT is a central point of the grammar.

 (
   block=block  -- more statements within {} braces
 |
   ASSERT assert_condition=expression
   (COLON assert_message=expression | 0) SEMICOLON
 |
   IF LPAREN if_condition=expression RPAREN if_statement=statement
   (ELSE else_statement=statement | 0)
   -- conflict: the old "dangling-else" problem...
   -- kdevelop-pg generates proper code matching as soon as possible.
 |
   FOR LPAREN for_control=for_control RPAREN for_statement=statement
 |
   WHILE LPAREN while_condition=expression RPAREN while_statement=statement
 |
   DO do_while_statement=statement
   WHILE LPAREN do_while_condition=expression RPAREN SEMICOLON
 |
   TRY try_block=block !(#handler=try_handler)
   (FINALLY finally_block=block | 0)
 |
   SWITCH LPAREN switch_expression=expression RPAREN
   LBRACE !(#switch_cases=switch_statements_group) RBRACE
 |
   SYNCHRONIZED LPAREN synchronized_locked_type=expression RPAREN
   synchronized_block=block
 |
   return_token=RETURN (return_expression=expression | 0) SEMICOLON
 |
   THROW throw_exception=expression SEMICOLON
 |
   break_token=BREAK (break_label=IDENTIFIER | 0) SEMICOLON
 |
   continue_token=CONTINUE (continue_label=IDENTIFIER | 0) SEMICOLON
 |
   SEMICOLON
 |
   ?[: LA(2).kind == Token_COLON :]
   labeled_statement_identifier=IDENTIFIER COLON labeled_statement=statement
 |
   expression_statement=expression SEMICOLON  -- method call, assignment, etc.
 )
-> statement ;;


-- TRY HANDLERs are the catch(...) {...} constructs in try/catch blocks.

   CATCH LPAREN exception_parameter=parameter_declaration RPAREN
   catch_block=block
-> try_handler ;;


-- A group of SWITCH STATEMENTS are any number of "case x:" or "default:"
-- labels followed by a list of statements.

   #case=switch_case !(#case=switch_case)
   !(#statement=statement)
-> switch_statements_group ;;

   ( token=CASE expression=expression | token=DEFAULT ) COLON
-> switch_case ;;


-- The FOR CONTROL is the three statements inside the for(...) parentheses.
-- It has two problematic conflicts in it, each of which is only solvable
-- with LL(k). The first one is here in for_control where we have a conflict
-- between for_clause (the for_init part of it) and the parameter_declaration
-- in the for-each clause.

 (
   -- The traditional version, part of a so-called "basic for statement".
   -- ?[: LA( for_clause_lookahead ) :] -- not supported by kdev-pg
   for_clause
 |
   -- The for-each version, part of a so-called "enhanced for statement".
   -- It looks for example like for( i : myList.values() ) {...}.
   for_each_parameter=parameter_declaration
   COLON iterable_expression=expression
 )
-> for_control ;;

--    (for_init | 0) SEMICOLON
-- -> for_clause_lookahead ;;  -- only use for lookaheads!

   (for_init=for_init                         | 0) SEMICOLON   -- "int i = 0;"
   (for_condition=expression                  | 0) SEMICOLON   -- "i < size;"
   (#for_update_expression=expression @ COMMA | 0)             -- "i++"
-> for_clause ;;

-- Here is the for statement's second problematic conflict:
-- Can't decide between variable_declaration and expression.
-- Only solvable with LL(k), too.

   -- ?[: LA( variable_declaration ) :] -- not supported by kdev-pg
   local_variable_declaration=variable_declaration
 | #expression=expression @ COMMA
-> for_init ;;




-- EXPRESSIONs
-- Note that most of these expressions follow the pattern
--   thisLevelExpression :
--     nextHigherPrecedenceExpression @ OPERATOR
-- which is a standard recursive definition for a parsing an expression.
-- The operators in java have the following precedences:
--  lowest  (13)  = *= /= %= += -= <<= >>= >>>= &= ^= |=
--          (12)  ?:
--          (11)  ||
--          (10)  &&
--          ( 9)  |
--          ( 8)  ^
--          ( 7)  &
--          ( 6)  == !=
--          ( 5)  < <= > >=
--          ( 4)  << >>
--          ( 3)  +(binary) -(binary)
--          ( 2)  * / %
--          ( 1)  ++ -- +(unary) -(unary)  ~  !  (type)
--                []   () (method call)  . (dot -- identifier qualification)
--                new   ()  (explicit parenthesis)
--
-- the last two are not usually on a precedence chart; they are put in
-- to point out that new has a higher precedence than '.', so you
-- can validy use
--   new Frame().show()


   conditional_expression=conditional_expression
   (op=assignment_operator assignment_expression=expression | 0)
-> expression ;;

   token=ASSIGN
 | token=PLUSASSIGN
 | token=MINUSASSIGN
 | token=STARASSIGN
 | token=SLASHASSIGN
 | token=ANDASSIGN
 | token=ORASSIGN
 | token=XORASSIGN
 | token=REMAINDERASSIGN
 | token=LSHIFTASSIGN
 | token=RSIGNEDSHIFTASSIGN
 | token=RUNSIGNEDSHIFTASSIGN
-> assignment_operator ;;


   logical_or_expression=logical_or_expression
   (  QUESTION if_expression=expression
      COLON    else_expression=conditional_expression
    | 0
   )
-> conditional_expression ;;

   #expression=logical_and_expression @ LOG_OR
-> logical_or_expression ;;

   #expression=bit_or_expression @ LOG_AND
-> logical_and_expression ;;

   #expression=bit_xor_expression @ BIT_OR
-> bit_or_expression ;;

   #expression=bit_and_expression @ XOR
-> bit_xor_expression ;;

   #expression=equality_expression @ BIT_AND
-> bit_and_expression ;;

   #expression=relational_expression @ (#op=equality_operator)
-> equality_expression ;;

   token=EQUAL
 | token=NOTEQUAL
-> equality_operator ;;

   #expression=shift_expression
   (  !(#op=relational_operator #expression=shift_expression)
    | INSTANCEOF instance_of_type=type_specification
   )
-> relational_expression ;;

   token=LESSTHAN
 | token=GREATERTHAN
 | token=LESSEQUAL
 | token=GREATEREQUAL
-> relational_operator ;;

   #expression=additive_expression @ (#op=shift_operator)
-> shift_expression ;;

   token=LSHIFT
 | token=RSIGNEDSHIFT
 | token=RUNSIGNEDSHIFT
-> shift_operator ;;

   #expression=multiplicative_expression @ (#op=additive_operator)
-> additive_expression ;;

   token=PLUS
 | token=MINUS
-> additive_operator ;;

   #expression=unary_expression @ (#op=multiplicative_operator)
-> multiplicative_expression ;;

   token=STAR
 | token=SLASH
 | token=REMAINDER
-> multiplicative_operator ;;

   INCREMENT incremented_expression=unary_expression
 | DECREMENT decremented_expression=unary_expression
 | MINUS unary_minus_expression=unary_expression
 | PLUS  unary_plus_expression=unary_expression
 | other_expression=unary_expression_not_plusminus
-> unary_expression ;;


-- So, up till now this was the easy stuff. Here come the most sincere
-- conflicts in the grammar, it seems they can't be solved with LL(1).
-- The conflict in this rule is the ambiguity between type casts
-- (which can also be arbitrary class names within parentheses)
-- and primary_expressions, which can also be a lot of things.
-- The solution would be to LL(k) look ahead if it's a cast expression,
-- if yes, then follow that path, if not, let it be a primary_expression.

   TILDE bitwise_not_expression=unary_expression
 | BANG  logical_not_expression=unary_expression
-- | -- ?[: LA( cast_expression_lookahead ) :]  -- not supported by kdev-pg
--   cast_expression  -- conflict, only solvable with LL(k).
 | primary_expression=primary_expression !(#postfix_operator=postfix_operator)
 | cast_expression=cast_expression  -- when solved, take the one before primary_expression
-> unary_expression_not_plusminus ;;


--    LPAREN type_specification RPAREN
-- -> cast_expression_lookahead ;;  -- only use for lookaheads!

   LPAREN
   (  builtin_type_specification=builtin_type_specification RPAREN
      builtin_casted_expression=unary_expression
    |
      class_type_specification=class_type_specification RPAREN
      class_casted_expression=unary_expression_not_plusminus
   )
-> cast_expression ;;

   token=INCREMENT
 | token=DECREMENT
-> postfix_operator ;;


-- PRIMARY EXPRESSIONs: qualified names, array expressions,
--                      method invocation, post increment/decrement

   primary_atom=primary_atom !(#selector=primary_selector)
-> primary_expression ;;

-- SELECTORs appended to a primary atom can provide access to ".this" or
-- ".super", create classes with ".new ClassName(...)", call methods with
-- ".methodName(...)", access a member variable with ".variableName",
-- the expression's class info with ".class" and access arrays with "[...]".

 (
   DOT
   (  class_token_dot=CLASS
    | this_token=THIS
    | new_expression=new_expression
    | ?[: LA(2).kind != Token_LPAREN :]  -- member variable access (no method call)
      variable_name=IDENTIFIER
    |
      -- method calls (including the "super" ones) may have type arguments
      (type_arguments=non_wildcard_type_arguments | 0)
      (  super_token=SUPER super_suffix=super_suffix
       | method_name=IDENTIFIER
         LPAREN method_arguments=argument_list RPAREN
      )
   )
 |
   LBRACKET array_index_expression=expression RBRACKET
 )
-> primary_selector ;;


-- SUPER SUFFIX: a call to either a constructor, a method or
--               a member variable of the super class.

 (
   LPAREN constructor_arguments=argument_list RPAREN  -- constructor call
 |
   DOT  -- member access
   (  ?[: LA(2).kind != Token_LPAREN :]  -- member variable access (no method call)
      variable_name=IDENTIFIER
    |
      -- method access (looks like super.methodName(...) in the end)
      (type_arguments=non_wildcard_type_arguments | 0)
      method_name=IDENTIFIER
      (LPAREN method_arguments=argument_list RPAREN | 0)
   )
 )
-> super_suffix ;;



-- PRIMARY ATOM: the basic element of a primary expression,
--               and expressions in general

 (
   builtin_type=builtin_type_specification
   DOT class_token_suffix=CLASS   -- things like int.class or int[].class
 |
   literal=literal
 |
   new_expression=new_expression
 |
   LPAREN parenthesis_expression=expression RPAREN
 |
   this_token=THIS
   (LPAREN this_constructor_arguments=argument_list RPAREN | 0)
 |
   super_token=SUPER super_suffix=super_suffix
 |
   -- type names (normal) - either pure, as method or like bla[][].class
   #identifier=IDENTIFIER @ DOT  -- identifier without type arguments
   (  LPAREN method_arguments=argument_list RPAREN
    | declarator_brackets=optional_declarator_brackets DOT CLASS
    | 0
   )
 |
   -- generic method invocation with type arguments:
   type_arguments=non_wildcard_type_arguments
   (  super_token=SUPER super_suffix=super_suffix
    | this_token=THIS LPAREN this_constructor_arguments=argument_list RPAREN
    | #identifier=IDENTIFIER @ DOT
      LPAREN method_arguments=argument_list RPAREN
   )
 )
-> primary_atom ;;


-- NEW EXPRESSIONs are allocations of new types or arrays.

   NEW
   (type_arguments=non_wildcard_type_arguments | 0)
   type=type_specification_noarray
   (  LPAREN class_constructor_arguments=argument_list
      (class_body=class_body | 0)
    |
      array_creator_rest=array_creator_rest
   )
-> new_expression ;;

-- This array creator rest can be either
-- a.) empty brackets with an optional initializer (e.g. "[][]{exp,exp}") or
-- b.) at least one filled bracket, afterwards any amount of empty ones

 (
   ?[: LA(2).kind == Token_RBRACKET :]
   mandatory_declarator_brackets=mandatory_declarator_brackets
   array_initializer=variable_array_initializer
 |
   LBRACKET #index_expression=expression RBRACKET
   !(  ?[: LA(2).kind != Token_RBRACKET :]
       LBRACKET #index_expression=expression RBRACKET
     |
       optional_declarator_brackets=optional_declarator_brackets
       [: break; :]  -- no more filled brackets after the empty ones
    )
 )
-> array_creator_rest ;;



-- QUIT COPYING TO THE ORIGINAL FILE HERE



-- An ARGUMENT LIST mostly occurs in method calls, always between an LPAREN
-- and an RPAREN, to specify the values for the called method/initializer/etc.

   #expression=expression @ COMMA
 | 0
-> argument_list ;;




-- MODIFIERs for Java classes, interfaces, class/instance vars and methods.
-- Sometimes not all of them are valid, but that has to be checked manually
-- after running the parser. The ANTLR grammar also does it this way.

   token=PRIVATE
 | token=PUBLIC
 | token=PROTECTED
 | token=STATIC
 | token=TRANSIENT
 | token=FINAL
 | token=ABSTRACT
 | token=NATIVE
 -- Neither in the Java spec nor in the JavaCC grammar, just in the ANTLR one:
 -- | token=THREADSAFE
 | token=SYNCHRONIZED
 | token=VOLATILE
 | token=STRICTFP
 -- A modifier may be any annotation (e.g. @bla), but not @interface.
 -- This condition resolves the conflict between modifiers
 -- and annotation type declarations:
 | ?[: yytoken == Token_AT && LA(2).kind != Token_INTERFACE :]
    annotation=annotation
-> modifier ;;


-- LITERALs: fixed, constant values that don't have to be evaluated.

   token=FALSE
 | token=TRUE
 | token=NULL
 | token=INTEGER_LITERAL
 | token=FLOATING_POINT_LITERAL
 | token=CHARACTER_LITERAL
 | token=STRING_LITERAL
-> literal ;;





--
-- Appendix 1: stubs, so that the (incomplete) parser can be created
--

   STUB_A -> class_declaration ;;
   STUB_B -> enum_declaration ;;
   STUB_C -> interface_declaration ;;
   STUB_D -> annotation_type_declaration ;;
   STUB_E -> for_clause ;;


--
-- Appendix 2: token list, so they will be defined in java.h
--

   ( keyword | seperator | operator | literal | IDENTIFIER )
-> token ;;

   (  ABSTRACT | ASSERT | BOOLEAN | BREAK | BYTE | CASE | CATCH | CHAR | CLASS | CONST | CONTINUE
    | DEFAULT | DO | DOUBLE | ELSE | ENUM | EXTENDS | FINAL | FINALLY | FLOAT | FOR | GOTO
    | IF | IMPLEMENTS | IMPORT | INSTANCEOF | INT | INTERFACE | LONG | NATIVE | NEW | PACKAGE
    | PRIVATE | PROTECTED | PUBLIC | RETURN | SHORT | STATIC | STRICTFP | SUPER | SWITCH
    | SYNCHRONIZED | THIS | THROW | THROWS | TRANSIENT | TRY | VOID | VOLATILE | WHILE )
-> keyword ;;

   ( LPAREN | RPAREN | LBRACE | RBRACE | LBRACKET | RBRACKET | SEMICOLON | COMMA | DOT | AT )
-> seperator ;;

   (  ASSIGN | LESSTHAN | BANG | TILDE | QUESTION | COLON | EQUAL | LESSEQUAL | GREATEREQUAL
    | NOTEQUAL | LOG_AND | LOG_OR | INCREMENT | DECREMENT | PLUS | MINUS | STAR | SLASH | BIT_AND
    | BIT_OR | XOR | REMAINDER | LSHIFT | PLUSASSIGN | MINUSASSIGN | STARASSIGN | SLASHASSIGN
    | ANDASSIGN | ORASSIGN | XORASSIGN | REMAINDERASSIGN | LSHIFTASSIGN | RSIGNEDSHIFTASSIGN
    | RUNSIGNEDSHIFTASSIGN | GREATERTHAN | RSIGNEDSHIFT | RUNSIGNEDSHIFT | TRIPLEDOT )
-> operator ;;
